##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""Implementation for schedule.
$Id: Schedule.py,v 1.1.2.2 2002/12/11 10:19:51 jack-e Exp $
"""
from __future__ import generators

import time
import sys

from copy import copy
from datetime import datetime, timedelta

from Persistence import Persistent
from Persistence.BTrees.OOBTree import OOBTree

import zLOG

from Transaction import get_transaction

from Zope.Security.SecurityManagement import newSecurityManager

from Zope.App.Traversing import getPhysicalPathString

from Zope.ComponentArchitecture import getService, queryUtility
from Zope.ContextWrapper import ContextMethod

from Zope.Exceptions.ExceptionFormatter import format_exception
from Zope.Exceptions import NotFoundError

from Zope.Event.ISubscriber import ISubscriber
from Zope.App.LifeCycle.ILifeCycle import IZopeStartupEvent, IZopeShutdownEvent

from Zope.App.OFS.Services.LocalEventService.PathSubscriber import PathSubscriber

from Zope.App.Tasks.IScheduleThreadManager import IScheduleThreadManager
from Zope.App.Tasks.ISchedule import ISchedule, IScheduleControl
from Zope.App.Tasks.ITask import ITask

# Implementation note:
#
# Internally, there's a BTree mapping time -> [tasks]


class Schedule(Persistent):

    __implements__ = (ISchedule, IScheduleControl, ISubscriber)

    def __init__(self):
        self.tasks = OOBTree()

    ## Implementation for ISchedule

    def schedule(self, task):
        tasks = self.tasks

        task_added = 0

        if ITask.isImplementedBy(task):
             time = task.time
             self.tasks[time] = self.tasks.get(time, ()) + (task, )
        else:
            # task does not Implement ITask
            raise TypeError
    schedule = ContextMethod(schedule)

    def unschedule(self, task):
        tasks = self.tasks

        if ITask.isImplementedBy(task):
            time = task.time

            if time in tasks:
                tasklist = tuple([ t for t in tasks[time] if not t is task ])
                if not tasklist:
                    del tasks[time]
                else:
                    tasks[time] = tasklist
            else:
                # task was not in queue
                raise ValueError
        else:
            # task does not implement ITask
            raise TypeError
    unschedule = ContextMethod(unschedule)

    # implementation for IScheduleControl

    def process(self, time, autocommit=0):
        
        for task_status in self.generatorProcess(time, autocommit):
            pass


    def generatorProcess(self, time, autocommit=0):
        # get Authentication Service
        auth_svc = getService(self, "Authentication")

        tasks = self.tasks

        # retrieve all tasklists up to time
        seq = tasks.items(None,time)

        # process tasks
        for time_key, tasklist in seq:

            del tasks[time_key]
            tasklist = [copy(t) for t in tasklist]

            for task in tasklist:
                print "processing task"
                # create SecurityContext 
                try:
                    principal = auth_svc.getPrincipal(task.principalId)
                except NotFoundError:
                    # unkown Principal -> don't execute Task
                    # XXX Log this
                    print "principal not found"
                    continue
                old_sec_manager = newSecurityManager(principal)
                try:

                    try:
                       errtask = task()
                    
                       if errtask is None:
                           # give task a chance to reschedule
                           newtask = task.schedule()
                           if newtask is not None:
                              self.schedule(newtask)
                       else:
                          self.schedule(errtask)

                       if autocommit:
                           get_transaction().commit()
                    except:
                       # prevent the scheduler from crashing 
                       # when bogus task raised exception
                       # need to Log Exception though
                       tb_text = ''.join(
                           format_exception(*sys.exc_info(), **{'as_html': 0}))
                       zLOG.LOG("Schedule",zLOG.ERROR,"Exception in Task: %s" % tb_text)
                       if autocommit:
                           get_transaction().abort()

                finally:
                    newSecurityManager(old_sec_manager)
            yield "Processed %s" % time_key
            
    process = ContextMethod(process)


    def getTaskTimes(self):
        return self.tasks.keys()

    def getTaskInfo(self, time):
        return ["%s, %s, %s" % (t.time, t.principalId, t.description) \
                for t in  self.tasks.get(time,())]
        



    ## Implementation for ISubscriber

    def notify(self, event):

        schedule_mgr = queryUtility(None, IScheduleThreadManager)
        if schedule_mgr is None:
            return
        schedule_path = getPhysicalPathString(self)
        
        if IZopeStartupEvent.isImplementedBy(event):
            schedule_mgr.register(schedule_path)
        elif IZopeShutdownEvent.isImplementedBy(event):
            schedule_mgr.unregister(schedule_path)
    notify = ContextMethod(notify)


    currentlySubscribed = False

    def subscribe(self, channel=None):
        if self.currentlySubscribed:
            raise RuntimeError, "already subscribed; please unsubscribe first"
        channel = self._getChannel(channel)
        subscriber = PathSubscriber(self)
        channel.subscribe(subscriber,IZopeStartupEvent)
        channel.subscribe(subscriber,IZopeShutdownEvent)
        self.currentlySubscribed = True
    subscribe = ContextMethod(subscribe)

    def unsubscribe(self, channel=None):
        if not self.currentlySubscribed:
            raise RuntimeError, "already subscribed; please unsubscribe first"
        channel = self._getChannel(channel)
        subscriber = PathSubscriber(self)
        channel.unsubscribe(subscriber,IZopeStartupEvent)
        channel.unsubscribe(subscriber,IZopeShutdownEvent)
        self.currentlySubscribed = False
    unsubscribe = ContextMethod(unsubscribe)


    def isSubscribed(self):
        return self.currentlySubscribed

    def _getChannel(self, channel):
        if channel is None:
            channel = getService(self,"Events")
        return channel
    _getChannel = ContextMethod(_getChannel)

    
