##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""Implementation for ScheduleService.
$Id: $
"""
from __future__ import generators

import time
import sys

from copy import copy
from datetime import datetime, timedelta

from Persistence import Persistent
from Persistence.BTrees.OOBTree import OOBTree
from ZODB.POSException import ConflictError

import zLOG

from Transaction import get_transaction

from Zope.Security.SecurityManagement import newSecurityManager

from Zope.App.Traversing import traverse, getPhysicalPathString

from Zope.ComponentArchitecture import getService, getUtility
from Zope.Proxy.ProxyIntrospection import removeAllProxies
from Zope.Proxy.ContextWrapper import ContextWrapper
from Zope.ContextWrapper import ContextMethod

from Zope.Exceptions.ExceptionFormatter import format_exception
from Zope.Exceptions import NotFoundError

from Zope.Event.ISubscriber import ISubscriber

from Zope.App.ZopePublication.ZopePublication import ZopePublication
from Zope.App.OFS.Services.LocalEventService.PathSubscriber import PathSubscriber

from Zope.App.LifeCycle.ILifeCycle import IZopeStartupEvent, IZopeShutdownEvent

from Zope.App.Tasks.ManagedThread import ManagedThread
from Zope.App.Tasks.ManagedThread import THREAD_WORKING, THREAD_SLEEPING, \
     THREAD_EXCEPTION, THREAD_FINISHED
from Zope.App.Tasks.IThreadManager import IThreadManager
from Zope.App.Tasks.ISchedule import ISchedule, IScheduleControl
from Zope.App.Tasks.ITask import ITask

from Zope.App.LifeCycle.ILifeCycle import IZopeStartupEvent, IZopeShutdownEvent

from Zope.App.Tasks.IScheduleThreadManager import IScheduleThreadManager




_min_timedelta = timedelta(0, 0, 1000)

def timedelta2seconds(delta):
    return delta.days*86400 + delta.seconds + float(delta.microseconds)  / 1000000


# what needs to be added to make this secure ??
# ProxyFactory ?
class MinimalPublication:

    root_name = ZopePublication.root_name

    def __init__(self, db):
        self.db = db
        self._conn = None

    def getApplication(self):
        self._conn = conn = self.db.open()
        root = conn.root()
        app = root.get(self.root_name, None)
        if app is None:
            raise SystemError, "Zope Application Not Found"        
        return app

    def cleanup(self):
        self._conn.close()




class ScheduleThread(ManagedThread):

    def __init__(self, global_subscribe=False):
        super(ScheduleThread, self).__init__()

        self.schedule_times = []
        self.schedule_next = datetime.now()
        
        self.global_subscribe = global_subscribe


    def doWork(self, schedule_path, poll_delay=3600):


        pub = MinimalPublication(self._zodb)
        app = pub.getApplication()
        schedule = traverse(app, schedule_path)
        self.updateScheduleTimes(schedule)
        pub.cleanup()
        
        while self.running:
        
            if self.schedule_times:
                now = datetime.now()
                if (self.schedule_next - now) <= _min_timedelta:
                    print "processing schedule %s (%s)" % (schedule_path, self.schedule_next)
                    
                    pub = MinimalPublication(self._zodb)
                    app = pub.getApplication()
                    get_transaction().begin()
                    schedule = traverse(app, schedule_path)
                    
                    for task_status in schedule.generatorProcess(now, autocommit=1):
                        print "tasklist processed: %s" % task_status
                        yield THREAD_WORKING, pub.cleanup

                    self.updateScheduleTimes(schedule)
                    get_transaction().commit()
                    pub.cleanup()
                    
                    while self.schedule_next < now and self.schedule_times:
                        self.schedule_next = self.schedule_times.pop(0)

                now = datetime.now()
                delta = self.schedule_next - now
                if delta > _min_timedelta:
                    sleep_delay = min(poll_delay, timedelta2seconds(delta))
                    print "ScheduleThread %s sleeping (%f sec)" % (schedule_path, sleep_delay)
                    yield THREAD_SLEEPING, sleep_delay
            else:
                pub = MinimalPublication(self._zodb)
                app = pub.getApplication()
                schedule = traverse(app, schedule_path)
                self.updateScheduleTimes(schedule)
                pub.cleanup()
                if not self.schedule_times:
                    print "ScheduleThread %s sleeping (%f sec) (no jobs)" % (schedule_path, poll_delay)
                    yield THREAD_SLEEPING, poll_delay
               
                
    #doWork = ContextMethod(doWork)

    def updateScheduleTimes(self,schedule):
        # poll schedule_times
        schedule_times = schedule.getTaskTimes()
        times = self.schedule_times
        
        times.append(self.schedule_next)
        times.extend([copy(t) for t in schedule_times if t])
        times.sort()
        
        if times:
            self.schedule_next = times.pop(0)
        else:
            # XXX not really a solution
            self.schedule_next = datetime.now()

            
class ScheduleThreadManager:
    """ScheduleThreadManager."""

    __implements__ = IScheduleThreadManager

    startScheduleThreads = True


    def __init__(self):
        self._schedules = {}


    def register(self, location, poll=5.0):
        if not location in self._schedules.keys():
            if self.startScheduleThreads:
                thread_manager = getUtility(None, IThreadManager)
                thread_object = ScheduleThread()
                thread_manager.add("ScheduleThread@%s" % location,
                                     thread_object, location, poll)
            else:
                thread_object = None
            self._schedules[location] = thread_object
        else:
            raise ValueError, "Schedule already registered."

    def unregister(self, location):
        if self._schedules.has_key(location):
            if self.startScheduleThreads:
                thread_object = self._schedules[location]
                if thread_object is not None:
                    thread_object.stop()
            del self._schedules[location]
        else:
            raise ValueError, "Schedule not registered."


    def listScheduleThreads(self):
        return self._schedules.keys()

    def getScheduleThread(self, location):
        return self._schedules[location]

ScheduleThreadManagerUtility = ScheduleThreadManager()
